@@ -1,6 +1,6 @@
 /*-------------------------------------------------------------------------*/
 /* colordefs.h 								   */
-/* Copyright (c) 1998  Tim Edwards, Johns Hopkins University       	   */
+/* Copyright (c) 2000  Tim Edwards, Johns Hopkins University       	   */
 /*-------------------------------------------------------------------------*/
 
 /*-------------------------------------------------------------------------*/
@@ -9,7 +9,7 @@
 /* And ApplicationDataPtr in xcircuit.h					   */
 /*-------------------------------------------------------------------------*/
 
-#define NUMBER_OF_COLORS	10
+#define NUMBER_OF_COLORS	11
 
 #define BACKGROUND	appcolors[0]
 #define FOREGROUND	appcolors[1]
@@ -21,8 +21,11 @@
 #define OFFBUTTONCOLOR	appcolors[7]
 #define AUXCOLOR	appcolors[8]
 #define BARCOLOR	appcolors[9]
+#define BBOXCOLOR	appcolors[10]
 
 #define DEFAULTCOLOR	-1		/* Inherits color of parent */
 #define DOFORALL	-2		/* All elements inherit same color */
+#define DOSUBSTRING	-3		/* Only selected substring drawn */
+					/* in the SELECTCOLOR		*/
 
 /*-------------------------------------------------------------------------*/
