@@ -1,6 +1,6 @@
 /*-------------------------------------------------------------------------*/
 /* elements.c --- xcircuit routines for creating basic on-screen elements  */
-/* Copyright (c) 1998  Tim Edwards, Johns Hopkins University        	   */
+/* Copyright (c) 2000  Tim Edwards, Johns Hopkins University        	   */
 /*-------------------------------------------------------------------------*/
 
 /*-------------------------------------------------------------------------*/
@@ -25,24 +25,24 @@
 #include "cursors.h"
 #include "colordefs.h"
 #include "xcircuit.h"
-#include "alpha2.h"
 
 /*-------------------------------------------------------------------------*/
 /* Global Variable definitions						   */
 /*-------------------------------------------------------------------------*/
 
-extern short eventmode;	/* keep track of the mode the screen is in */
-extern short textpos;	/* keep track of the cursor position in text */
-extern Display	*dpy;   /* Works well to make this globally accessible */
+extern short eventmode;		/* keep track of the mode the screen is in */
+extern short textpos, textend;	/* keep track of the cursor position in text */
+extern Display	*dpy;   	/* Works well to make this globally accessible */
 extern Window win;
 extern int *appcolors;
 extern Cursor	appcursors[NUM_CURSORS];
 extern Clientdata areastruct;
+extern Globaldata xobjs;
 extern Widget	top;
-extern char **fonts;
+extern fontinfo *fonts;
 extern short fontcount;
 extern short attachto;
-extern char _STR[150];
+extern char  _STR[150], _STR2[250];
 extern char *labelbuf;
 
 #define RADFAC 0.0174532925199   /* pi / 180 */
@@ -53,21 +53,65 @@ extern double atan2();
 /*------------------------------------------------------------------------*/
 
 short savedir = 0;	/* for rhomboid/manhattan editing */
-short extmode = 0;	/* for text escape sequences */
 char extchar[20];
 double saveratio;
+unsigned char texttype;
 
 /*------------------------------------------------------------------------*/
 /* Declarations of functions defined externally to elements.c	 	  */
 /*------------------------------------------------------------------------*/
 
+extern TextExtents ULength();
 extern void undrawtext(), redrawtext();
 extern XtEventHandler trackspline(), trackarc(), trackbox();
+objectptr finddot();
 
 #ifdef SCHEMA
 extern labelptr findotherpin();
 #endif
 
+/*--------------------------------------*/
+/* Draw a dot at the current point.	*/
+/*--------------------------------------*/
+
+void drawdot(int xpos, int ypos)
+{
+   arcptr *newarc;
+   objinstptr *newdot;
+   objectptr dotobj;
+   
+   /* Find the object "dot" in the builtin library, or else use an arc */
+   
+   if ((dotobj = finddot()) != (objectptr)NULL) {
+      NEW_OBJINST(newdot, objectdata);
+      objectdata->parts++;
+      (*newdot)->position.x = xpos;
+      (*newdot)->position.y = ypos;
+      (*newdot)->rotation = 1;
+      (*newdot)->scale = 1.0;
+      (*newdot)->thisobject = dotobj;
+      (*newdot)->color = DEFAULTCOLOR;
+#ifdef SCHEMA
+      (*newdot)->params = NULL;
+      (*newdot)->num_params = (uchar)0;
+#endif
+   }
+   else {
+      NEW_ARC(newarc, objectdata);
+      objectdata->parts++;
+      (*newarc)->position.x = xpos;
+      (*newarc)->position.y = ypos;
+      (*newarc)->angle1 = 0;
+      (*newarc)->angle2 = 360;
+      (*newarc)->width = 1.0;
+      (*newarc)->radius = 6;
+      (*newarc)->yaxis = 6;
+      (*newarc)->style = FILLED | FILLSOLID | NOBORDER;
+      (*newarc)->color = DEFAULTCOLOR;
+      calcarc(*newarc);
+   }
+}
+
 /*----------------------*/
 /* Start spline mode.   */
 /*----------------------*/
@@ -93,10 +137,8 @@ void starttext(button, mode, calldata)
 {
    if (eventmode == NORMAL_MODE) {
       Wprintf("Click button1 for text.");
-      if (mode == 0)
-         eventmode = TEXT0_MODE;
-      else
-	 eventmode = TEXT1_MODE;	/* Pinlabel */
+      texttype = (unsigned char)mode;
+      eventmode = TEXT1_MODE;
    }
 }
 
@@ -117,6 +159,9 @@ void textbutton(w, dopin, event)
    XDefineCursor(dpy, win, TEXTPTR);
    Wprintf("Click to end or cancel.");
 
+   if (fontcount == 0)
+      Wprintf("Warning:  No fonts available!");
+
    NEW_LABEL(newlabel, objectdata);
    areastruct.editpart = objectdata->parts;
 
@@ -132,9 +177,10 @@ void textbutton(w, dopin, event)
 
 #ifdef SCHEMA
    (*newlabel)->pin = False;
-   if (dopin == True && areastruct.schemon) {
-      (*newlabel)->pin = LOCAL;
-      (*newlabel)->color = SNAPCOLOR;
+   if (areastruct.schemon) {
+      (*newlabel)->pin = dopin;
+      if (dopin == LOCAL) (*newlabel)->color = SNAPCOLOR;
+      else if (dopin == INFO) (*newlabel)->color = AUXCOLOR;
    }
 #endif
 
@@ -152,19 +198,27 @@ void textbutton(w, dopin, event)
    areastruct.origin.y = userpt.y;
 }
 
-/*----------------------------------*/
-/* Report on characters near cursor */
-/*----------------------------------*/
+/*----------------------------------------------------------------------*/
+/* charreport():  Report on characters near cursor.			*/
+/* Cursor position is effectively between (textpos - 1) and (textpos).	*/
+/* For purposes of printing non-printable character sequences, it is	*/
+/* assumed that routine labeltext() always places the edit position	*/
+/* ("textpos") after the END of the sequence (i.e., *(textpos - 2) =	*/
+/* TEXT_ESC).								*/
+/*----------------------------------------------------------------------*/
 
 static char *nonprint[] = {
 	"Subscript", "Superscript", "Normalscript", "Underline", "Overline",
-	"Noline", "Backspace", "Halfspace", "Quarterspace"};
+	"Noline", "Backspace", "Halfspace", "Quarterspace", "Special",
+	"N/A", "Scale Change", "Parameter (start)", "Parameter (end)",
+	"<Return>", "N/A", "Color Change", "N/A", "Kern", "Error"};
 
 void charreport(curlabel)
    labelptr curlabel;
 {
    char report[2][50];
-   uchar *curpos = curlabel->string + textpos;
+   uchar *pcurpos = curlabel->string + textpos;
+   uchar *ncurpos = pcurpos - 1;
    uchar *endpos;
 
    report[0][0] = '\0';
@@ -172,32 +226,58 @@ void charreport(curlabel)
 
    for (endpos = curlabel->string; *endpos != '\0'; endpos++);
 
-   curpos = curlabel->string + textpos;
-   if (curpos > curlabel->string) {
-      if ((int)(*(curpos - 2)) == TEXT_ESC) {
-	 int control = (int)(*(curpos - 1));
+   while (*ncurpos == NO_OP && ncurpos >= curlabel->string) ncurpos--;
+   if (ncurpos >= curlabel->string) {
+      if ((int)(*(ncurpos - 1)) == TEXT_ESC) {
+	 int control = (int)(*ncurpos);
 
          if (control >= FONT_START)
-            strcpy(report[0], fonts[control - FONT_START]);
-         else
+            strcpy(report[0], fonts[control - FONT_START].psname);
+         else if (control >= CONTROL_START)
 	    strcpy(report[0], nonprint[control - CONTROL_START]);
+	 else
+	    strcpy(report[0], "Error");
+      }
+      else if (*ncurpos == '\\') {
+	 report[0][0] = *ncurpos;
+	 report[0][1] = ' ';
+	 report[0][2] = '\0';
       }
       else  {
-	 report[0][0] = *(curpos - 1);
+	 report[0][0] = *ncurpos;
 	 report[0][1] = '\0';
       }
    }
-   if (curpos < endpos) {
-      if ((int)(*curpos) == TEXT_ESC) {
-	 int control = (int)(*(curpos + 1));
+
+   while (*pcurpos == NO_OP && pcurpos < endpos) pcurpos++;
+   if (*pcurpos == TEXT_ESC) {
+      switch (*(pcurpos + 1)) {
+	 case FONT_SCALE:
+	    pcurpos += 2;
+	 case KERN:
+	    pcurpos++;
+	 case FONT_COLOR:
+	    pcurpos += 3;
+      }
+   }
+   if (pcurpos < endpos) {
+      if ((int)(*pcurpos) == TEXT_ESC) {
+	 int control = (int)(*(pcurpos + 1));
 
          if (control >= FONT_START)
-            strcpy(report[1], fonts[control - FONT_START]);
-         else
+            strcpy(report[1], fonts[control - FONT_START].psname);
+         else if (control >= CONTROL_START)
 	    strcpy(report[1], nonprint[control - CONTROL_START]);
+	 else
+	    strcpy(report[1], "Error");
+      }
+      else if (*pcurpos == '\\') {
+	 report[1][0] = *pcurpos;
+	 report[1][1] = ' ';
+	 report[1][2] = '\0';
       }
       else  {
-	 report[1][0] = *curpos;
+	 report[1][0] = *pcurpos;
 	 report[1][1] = '\0';
       }
    }
@@ -219,7 +299,7 @@ void labeltext(keypressed, keystate)
   int keystate;		     /* 1 = Control , 2 = keypressed is font number */
 {
    labelptr curlabel;
-   uchar *curpos, *endpos, *sptr;
+   uchar *curpos, *endpos, *sptr, *oldstr;
    char report[2][50];
    XPoint cursorpt;
    short tmplength, tmpheight, cfont, ncfont;
@@ -243,119 +323,75 @@ void labeltext(keypressed, keystate)
 	 break;
       }
 
-   if (extmode == 0) UDrawTLine(curlabel);
-
    /* Generate non-ascii extended character set characters */
 
    if (keypressed == XK_backslash) {
-      if (extmode == 0) {
-         extmode = 1;
-	 extchar[0] = extchar[1] = extchar[2] = ' ';
-	 extchar[3] = '\0';
-         Wprintf("Enter decimal code, symbol name, or '\\':");
+      if (keystate == 0) {
+         composefontlib(cfont);
+         startcatalog(NULL, Number(FONTLIB), NULL);
          return;
       }
-      else if (extmode == 1) extmode = 0;
+      else
+         keystate = 0;
    }
-   if (extmode > 0) {
-      if (extmode > 1 && (keypressed == XK_BackSpace || keypressed == XK_Delete)) {
-	 if (extmode != 5) {
-	    extmode--;
-	    _STR[strlen(_STR) - 1] = '\0';
-            Wprintf(_STR);
-	 }
-	 else {
-	    extmode = 1;
-            Wprintf("Enter decimal code, symbol name, or '\\':");
-	 }
-	 return;
+
+   UDrawTLine(curlabel);
+
+   if (keypressed == XK_BackSpace) keypressed = XK_Delete;
+
+   if (keypressed == XK_Delete && curpos > curlabel->string + 2) {
+      short k, skip = 1;
+
+      /* special case:  multiple characters generated */
+      if (textend > 0 && textend < textpos) {
+	 skip = textpos - textend;
       }
-      else if (extmode < 5 && keypressed >= XK_0 && keypressed <= XK_9) {
-         extmode++;
-         extchar[extmode - 2] = (char)keypressed;
-         sprintf(_STR, "Enter decimal code: %c%c%c", extchar[0], extchar[1], extchar[2]);
-         Wprintf(_STR);
-         if (extmode < 4) return;
-         else { 
-	    extmode = 0;
-	    keypressed = (extchar[0] - '0') * 100 + (extchar[1] - '0') * 10 +
-		 (extchar[2] - '0');
+      else {
+         while (*(curpos - skip) == NO_OP) skip++;
+         if (*(curpos - skip - 1) == TEXT_ESC) {
+	    switch (*(curpos - skip)) {
+	       case END_SCALE:
+	          skip += 2;
+	       case END_KERN:
+	          skip++;
+	       case END_COLOR:
+	          skip += 3;
+	    }
+	    if (*(curpos - skip - 2) == TEXT_ESC) skip++;
+	    skip++;
          }
       }
-      else if (keypressed >= 32 && keypressed < 128) {
-	 if (extmode == 1) extmode = 5;
-	 else if (extmode < 5) return;
-	 else extmode++;
-	 extchar[extmode - 5] = (char)keypressed;
-	 extchar[extmode - 4] = '\0';
-	 sprintf(_STR, "Enter symbol name: %s", extchar);
-	 Wprintf(_STR);
-	 return;
+
+      /* Don't delete any parameter boundaries---must use	*/
+      /* "unparameterize" command for that.			*/
+
+      for (k = 1; k <= skip; k++) {
+         if (*(curpos - k - 1) == TEXT_ESC && (*(curpos - k) == PARAM_START
+		    || *(curpos - k) == PARAM_END)) {
+	    skip = k - 1;
+	    break;
+         }
       }
-      else if (keypressed == XK_Return) {
-	 short i;
-	 extmode = 0;
-	 for (i = 0; i < SNAMES; i++)
-	    if (!strcmp(syms[i].name, extchar)) break;
-	 if (i == SNAMES) {	/* check for font name */
-	    for (i = 0; i < fontcount; i++)
-	       if (!strcmp(fonts[i], extchar)) break;
-	    if (i < fontcount)
-	       keypressed = i + FONT_START;
-	    else {
-	       sprintf(_STR, "\"%s\" is not a known symbol name.", extchar);
-	       Wprintf(_STR);
-	       return;
-	    }
-	 }
-	 else {
-	    keypressed = syms[i].code;
 
-	    /* if current font is inconsistent with symbol, change font     */
-	    /* Return to original font afterward.			    */
-
-	    /* If current font is type symbol but new character is not.     */
-
-	    if (cfont >= 11 && syms[i].font == 1) {
-		UDrawTLine(curlabel);
-		labeltext(0 + (cfont % 4) + FONT_START, 2);
-		labeltext(keypressed, 0);
-		labeltext(cfont + FONT_START, 2);
-		return;
-	    }
-
-	    /* If new character is type symbol but current font is not.     */
-	    /* (cfont & 0xfd) clears bit which declares font as weight BOLD */
-	    /* because we have no bold symbol fonts. . .		    */
-
-	    else if (cfont < 12 && syms[i].font == 2) {
-		UDrawTLine(curlabel);
-		labeltext(12 + ((cfont % 4) & 0xfd) + FONT_START, 2);
-		labeltext(keypressed, 0);
-		labeltext(cfont + FONT_START, 2);
-		return;
-	    }
-	 }
+      if (textpos - skip >= 2) {
+         undrawtext(curlabel);
+         textpos -= skip;
+         for (; curpos <= endpos; curpos++) *(curpos - skip) = *curpos;
       }
-      else return;
-   }
-
-   else if (keypressed == XK_BackSpace)
-      keypressed = XK_Delete;
-
-   if (keypressed == XK_Delete && curpos > curlabel->string + 2){
-      short skip = (*(curpos - 2) == TEXT_ESC) ? 2 : 1;
-      undrawtext(curlabel);
-      textpos -= skip;
-      for (; curpos <= endpos; curpos++) *(curpos-skip) = *curpos;
+      textend = 0;
    }
    else if (keypressed == XK_Return) {
       Boolean hasstuff = False;   /* Check for null string */
+      Boolean do_all = True;
       uchar *tmppos;
-      Widget menuitem;
+      Widget w;
 
       for (tmppos = curlabel->string; tmppos < endpos; tmppos++) {
-	 if ((short)(*tmppos) == TEXT_ESC) tmppos++;
+	 if ((short)(*tmppos) == TEXT_ESC) {
+	    tmppos++;
+	    /* An empty parameter is still considered "stuff" */
+	    if ((short)(*tmppos) == PARAM_START) hasstuff = True;
+	 }
 	 else hasstuff = True;
       }
       XDefineCursor(dpy, win, CROSS);
@@ -404,47 +440,116 @@ void labeltext(keypressed, keystate)
       else eventmode = NORMAL_MODE;
 
       setdefaultfontmarks();
+      setcolormark(areastruct.color);
+      if (labelbuf != NULL) {
+         do_all = resolveparams(objectdata, labelbuf, curlabel);
+	 free(labelbuf);
+	 labelbuf = NULL;
+	 if (do_all >= 0)
+            calcbboxparam(objectdata, do_all);
+	 else /* do_all == -1 */
+	    calcbbox(objectdata);
+#ifdef SCHEMA
+	 setobjecttype(objectdata);
+#endif
+      }
+      else
+         calcbbox(objectdata);
       return;
    }
    else if (keypressed == XK_Right && curpos != endpos) {
+      uchar tchr;
       textpos++;
-      if (*(curlabel->string + textpos - 1) == TEXT_ESC) textpos++;
+      while (*(curlabel->string + textpos) == NO_OP) textpos++;
+      if (*(curlabel->string + textpos - 1) == TEXT_ESC) {
+	 textpos++;
+         /* Quick & dirty way to skip 7 for FONT_SCALE, 5 for KERN, 4 for FONT_COLOR */
+         switch (*(curlabel->string + textpos - 1)) {
+            case FONT_SCALE:
+	       textpos += 2;
+            case KERN:
+	       textpos++;
+            case FONT_COLOR:
+	       textpos += 3;
+	 }
+      }
    }
    else if (keypressed == XK_Left && textpos > 2) {
       textpos--;
-      if (*(curlabel->string + textpos - 1) == TEXT_ESC) textpos--;
+      while (*(curlabel->string + textpos) == NO_OP) textpos--;
+      if (*(curlabel->string + textpos - 1) == TEXT_ESC) {
+         /* Tricky pass-through switch statement (no breaks!): see above */
+         switch (*(curlabel->string + textpos)) {
+	    case END_SCALE:
+	       textpos -= 2;
+	    case END_KERN:
+	       textpos--;
+	    case END_COLOR:
+	       textpos -= 3;
+	 }
+	 textpos--;
+      }
    }
-   else if (keypressed == XK_A && keystate == 1) textpos = 2;
-   else if (keypressed == XK_E && keystate == 1) textpos = (int)(endpos -
-	 curlabel->string);
+
+   /* Ctrl-A and Ctrl-E have been replaced by Home and End, thus */
+   /* freeing up the control characters for ISO-Latin bindings.  */
+
+   else if (keypressed == XK_Home) textpos = 2;
+   else if (keypressed == XK_End) textpos = (int)(endpos - curlabel->string);
 
    /* Write a font designator or other control into the string */
 
+   oldstr = curlabel->string;
    if (keystate == 2) {
       short slen = endpos - curlabel->string;
+      short pad = 2;
+
+      switch(keypressed) {
+	 case FONT_SCALE:
+	    pad += 2;
+	 case KERN:
+	    pad++;
+	 case FONT_COLOR:
+	    pad += 3;
+      }
 
       curlabel->string = (uchar *) realloc(curlabel->string,
-         (slen + 3) * sizeof(uchar));
+         (slen + pad + 1) * sizeof(uchar));
       for (endpos = curlabel->string; *endpos != '\0'; endpos++);
       curpos = curlabel->string + textpos;
 
-      /* if center or right justified, must erase first. */
-      if ((curlabel->justify & NOTLEFT) || (curpos != endpos))
-         undrawtext(curlabel);
-      for (curpos = endpos + 2; curpos > curlabel->string + textpos; curpos--)
-         *curpos = *(curpos-2);
+      /* erase first before redrawing unless the string is empty */
+      undrawtext(curlabel);
+      for (curpos = endpos + pad; curpos > curlabel->string + textpos; curpos--)
+         *curpos = *(curpos - pad);
       *(curpos++) = (uchar)TEXT_ESC;
       *curpos = (uchar)keypressed;
-      textpos += 2;
-      *(curlabel->string + slen + 2) = '\0';
+      switch (keypressed) {
+	 case FONT_SCALE: case KERN: case FONT_COLOR:
+	    strncpy(curpos + 1, _STR2, pad - 4);
+	    *(curpos + pad - 3) = (uchar)TEXT_ESC;
+	    switch (keypressed) {
+	       case FONT_SCALE:
+		  *(curpos + pad - 2) = (uchar)END_SCALE;
+		  break;
+	       case KERN:
+		  *(curpos + pad - 2) = (uchar)END_KERN;
+		  break;
+	       case FONT_COLOR:
+		  *(curpos + pad - 2) = (uchar)END_COLOR;
+		  break;
+	    }
+	    break;
+      }
+      textpos += pad;
+      *(curlabel->string + slen + pad) = '\0'; /* redundant? */
    }
 
    if (keypressed > 0 && keypressed < 256 && keystate == 0) {
       short slen = endpos - curlabel->string;
 
-      /* if center or right justified, must erase first. */
-      if ((curlabel->justify & NOTLEFT) || (curpos != endpos))
-	 undrawtext(curlabel);
+      /* erase first. */
+      undrawtext(curlabel);
       for (curpos = endpos; curpos > curlabel->string + textpos; curpos--)
 	 *curpos = *(curpos-1);
       *curpos = (uchar)keypressed;
@@ -458,18 +563,24 @@ void labeltext(keypressed, keystate)
    if (symlabel != NULL) symlabel->string = curlabel->string;
 #endif
 
+   /* Keep track of parameter pointers when the string is reallocated */
+   if (curlabel->string != oldstr)
+      rehook(objectdata, oldstr, curlabel->string);
 
    if ((keypressed > 0 && keypressed < 256) || keypressed == XK_Delete ||
 	keypressed == XK_Right || keypressed == XK_Left || (keystate == 1
 	&& (keypressed == XK_A || keypressed == XK_B)) || keystate == 2) {
       char savchar = *(curlabel->string + textpos);
       short beglength;
+      TextExtents tmpext;
 
       *(curlabel->string + textpos) = '\0';
-      beglength = ULength(curlabel->string, curlabel->scale, 0);
+      tmpext = ULength(curlabel->string, curlabel->scale, 0, NULL);
+      beglength = tmpext.width;
       *(curlabel->string + textpos) = savchar;
 	
-      tmplength = ULength(curlabel->string, curlabel->scale, 0);
+      tmpext = ULength(curlabel->string, curlabel->scale, 0, NULL);
+      tmplength = tmpext.width;
       tmpheight = (short)(curlabel->scale * TEXTHEIGHT);
       areastruct.origin.x = curlabel->position.x + (curlabel->justify & NOTLEFT
 	  ? (curlabel->justify & RIGHT ? 0 : tmplength / 2) : tmplength)
@@ -477,12 +588,18 @@ void labeltext(keypressed, keystate)
       areastruct.origin.y = curlabel->position.y + (curlabel->justify &
 	  NOTBOTTOM ? (curlabel->justify & TOP ? -tmpheight : -tmpheight / 2)
 	  : 0);
+#ifdef SCHEMA
+      if (curlabel->pin)
+         pinadjust(curlabel->justify, &(areastruct.origin.x),
+		&(areastruct.origin.y), 1);
+#endif
       XcSetFunction(GXcopy);
       redrawtext(curlabel);
    }
    UDrawTLine(curlabel);
 
    charreport(curlabel);
+   textend = 0;
 
    /* find current font and adjust menubuttons as necessary */
 
@@ -495,6 +612,15 @@ void labeltext(keypressed, keystate)
    if (ncfont != cfont) setfontmarks(ncfont, -1);
 }
 
+/*-------------------------------------*/
+/* Initiate return from text edit mode */
+/*-------------------------------------*/
+
+void textreturn()
+{
+   labeltext(XK_Return, 0);
+}
+
 /*-------------------------------------*/
 /* Change the justification of a label */
 /*-------------------------------------*/
@@ -510,7 +636,7 @@ void rejustify(mode)
       curlabel = TOLABEL(EDITPART);
       UDrawTLine(curlabel);
       undrawtext(curlabel);
-      curlabel->justify = transjust[mode];
+      curlabel->justify = transjust[mode] | (curlabel->justify & FLIPINV);
       redrawtext(curlabel);
       UDrawTLine(curlabel);
 
@@ -522,7 +648,7 @@ void rejustify(mode)
 	 if (SELECTTYPE(tsel) == LABEL) {
 	    curlabel = SELTOLABEL(tsel);
 	    undrawtext(curlabel);
-	    curlabel->justify = transjust[mode];
+      	    curlabel->justify = transjust[mode] | (curlabel->justify & FLIPINV);
 	    redrawtext(curlabel);
 	 }
       }
@@ -555,9 +681,9 @@ void splinebutton(w, clientdata, event)
       (*newspline)->ctrl[j].x = userpt.x;
       (*newspline)->ctrl[j].y = userpt.y;
    }
-   (*newspline)->ctrl[1].x += areastruct.gridspace << 1;
-   (*newspline)->ctrl[2].x -= areastruct.gridspace << 1;
-   (*newspline)->width = 1.0;
+   (*newspline)->ctrl[1].x += (int)(xobjs.pagelist[areastruct.page]->gridspace / 2);
+   (*newspline)->ctrl[2].x -= (int)(xobjs.pagelist[areastruct.page]->gridspace / 2);
+   (*newspline)->width = areastruct.linewidth;
    areastruct.editcycle = 3;
    calcspline(*newspline);
 
@@ -584,12 +710,13 @@ XtEventHandler trackspline(w, clientdata, calldata)
    unsigned int nullui;
    splineptr    newspline;
 
-   newspline = TOSPLINE(EDITPART);
+   newspline = (eventmode == EPATH_MODE) ?
+               (splineptr)(*((*((pathptr *)EDITPART))->plist
+		+ areastruct.editsubpart)) : TOSPLINE(EDITPART);
    XQueryPointer(dpy, win, &nullwin, &nullwin, &nullint, &nullint, &xpos, &ypos,
 	&nullui);
    snap(xpos, ypos, &newpos);
    if (areastruct.save.x == newpos.x && areastruct.save.y == newpos.y) return;
-   printpos(newpos.x, newpos.y);
 
    UDrawEditSpline(areastruct.topobject, newspline);
 
@@ -604,6 +731,7 @@ XtEventHandler trackspline(w, clientdata, calldata)
    calcspline(newspline);
    UDrawEditSpline(areastruct.topobject, newspline);
 
+   printpos(newpos.x, newpos.y);
    areastruct.save.x = newpos.x;
    areastruct.save.y = newpos.y;
 }
@@ -642,7 +770,7 @@ void arcbutton(w, clientdata, event)
    snap(event->x, event->y, &userpt);
    (*newarc)->position.x = userpt.x;
    (*newarc)->position.y = userpt.y;
-   (*newarc)->width = 1.0;
+   (*newarc)->width = areastruct.linewidth;
    (*newarc)->radius = 0;
    (*newarc)->yaxis = 0;
    (*newarc)->angle1 = 0;
@@ -677,13 +805,14 @@ XtEventHandler trackarc(w, clientdata, calldata)
    arcptr      newarc;
    double adjrat;
 
-   newarc = TOARC(EDITPART);
+   newarc = (eventmode == EPATH_MODE) ?
+	    (arcptr)(*((*((pathptr *)EDITPART))->plist
+	    + areastruct.editsubpart)) : TOARC(EDITPART);
    XQueryPointer(dpy, win, &nullwin, &nullwin, &nullint, &nullint, &xpos, &ypos,
 	&nullui);
 
    snap(xpos, ypos, &newpos);
    if (areastruct.save.x == newpos.x && areastruct.save.y == newpos.y) return;
-   printpos(newpos.x, newpos.y);
 
    UDrawArc(areastruct.topobject, newarc);
    UDrawXLine(areastruct.save, newarc->position);
@@ -691,7 +820,8 @@ XtEventHandler trackarc(w, clientdata, calldata)
    if (areastruct.editcycle == 1 || areastruct.editcycle == 2) {
       float *angleptr, tmpang;
 
-      adjrat = (double)(abs(newarc->radius)) / (double)newarc->yaxis;
+      adjrat = (newarc->yaxis == 0) ? 1 :
+		(double)(abs(newarc->radius)) / (double)newarc->yaxis;
       angleptr = (areastruct.editcycle == 1) ? &newarc->angle1 : &newarc->angle2;
       tmpang = (float)(atan2((double)(newpos.y - newarc->position.y) * adjrat,
 	   (double)(newpos.x - newarc->position.x)) / RADFAC);
@@ -727,6 +857,7 @@ XtEventHandler trackarc(w, clientdata, calldata)
 
    UDrawArc(areastruct.topobject, newarc);
    UDrawXLine(newpos, newarc->position);
+   printpos(newpos.x, newpos.y);
 
    areastruct.save.x = newpos.x;
    areastruct.save.y = newpos.y;
@@ -764,7 +895,7 @@ void boxbutton(w, clientdata, event)
 
    (*newbox)->style = areastruct.style & ~UNCLOSED;
    (*newbox)->color = areastruct.color;
-   (*newbox)->width = 1.0;
+   (*newbox)->width = areastruct.linewidth;
    (*newbox)->number = 4;
    (*newbox)->points = (pointlist) malloc(4 * sizeof(XPoint));
    snap(event->x, event->y, &userpt);
@@ -804,7 +935,6 @@ XtEventHandler trackbox(w, clientdata, calldata)
 
    snap(xpos, ypos, &newpos);
    if (areastruct.save.x == newpos.x && areastruct.save.y == newpos.y) return;
-   printpos(newpos.x, newpos.y);
 
    UDrawPolygon(areastruct.topobject, newbox);
 
@@ -813,6 +943,7 @@ XtEventHandler trackbox(w, clientdata, calldata)
    pointptr++; pointptr->x = newpos.x;
    
    UDrawPolygon(areastruct.topobject, newbox);
+   printpos(newpos.x, newpos.y);
 
    areastruct.save.x = newpos.x;
    areastruct.save.y = newpos.y;
@@ -868,7 +999,7 @@ void startwire(userpt)
    (*newwire)->style = UNCLOSED | (areastruct.style & (DASHED | DOTTED));
    (*newwire)->color = areastruct.color;
    (*newwire)->number = 2;
-   (*newwire)->width = 1.0;
+   (*newwire)->width = areastruct.linewidth;
    (*newwire)->points = (pointlist) malloc(2 * sizeof(XPoint));
    pointptr = (*newwire)->points;
    pointptr->x = (pointptr + 1)->x = areastruct.save.x = userpt.x;
@@ -975,12 +1106,13 @@ void trackpoly(w, clientdata, calldata)
    unsigned int   nullui;
    polyptr      newpoly;
 
-   newpoly = TOPOLY(EDITPART);
+   newpoly = (eventmode == EPATH_MODE) ?
+	     (polyptr)(*((*((pathptr *)EDITPART))->plist
+	     + areastruct.editsubpart)) : TOPOLY(EDITPART);
    XQueryPointer(dpy, win, &nullwin, &nullwin, &nullint, &nullint, &xpos, &ypos,
 	&nullui);
    snap(xpos, ypos, &newpos);
    if (areastruct.save.x == newpos.x && areastruct.save.y == newpos.y) return;
-   printpos(newpos.x, newpos.y);
 
    UDrawPolygon(areastruct.topobject, newpoly);
 
@@ -1024,6 +1156,7 @@ void trackpoly(w, clientdata, calldata)
    }
 
    UDrawPolygon(areastruct.topobject, newpoly);
+   printpos(newpos.x, newpos.y);
    areastruct.save.x = newpos.x;
    areastruct.save.y = newpos.y;
 }
@@ -1232,7 +1365,8 @@ void join()
    pointlist newpoints;
    short     *scount, *sptr, *sptr2, *direc, *order;
    short     ordered, startpt = 0;
-   short     numpolys, numpoints, polytype, polycolor;
+   short     numpolys, numpoints, polytype;
+   int	     polycolor;
    float     polywidth;
    XPoint    *testpoint, *testpoint2, *begpoint, *endpoint, arcpoint[4];
    XPoint    *begpoint2, *endpoint2;
@@ -1500,7 +1634,7 @@ void wirebutton(w, clientdata, event)
    tpoint->x = userpt.x;
    tpoint->y = userpt.y;
 
-   /* prevent length zero wires */
+   /* back up one point; prevent length zero wires */
    if ((event->button == Button3) || ((tpoint - 1)->x == userpt.x &&
 	   (tpoint - 1)->y == userpt.y)) {
       if (newwire->number <= 2) {
@@ -1513,7 +1647,6 @@ void wirebutton(w, clientdata, event)
          if (--newwire->number == 2) newwire->style = UNCLOSED |
    		(areastruct.style & (DASHED | DOTTED));
       }
-      if (event->button == Button3) eventmode = NORMAL_MODE;
    }
 
    if (newwire && event->button == Button1) {
@@ -1524,7 +1657,7 @@ void wirebutton(w, clientdata, event)
       tpoint->x = userpt.x;
       tpoint->y = userpt.y;
    }
-   else
+   else if (newwire == NULL || event->button != Button3)
       XtRemoveEventHandler(areastruct.area, PointerMotionMask, False,
          (XtEventHandler)trackwire, NULL);
 
@@ -1535,6 +1668,8 @@ void wirebutton(w, clientdata, event)
 	 objectdata->parts++;
       }
       UDrawPolygon(areastruct.topobject, newwire);
+      if (event->button == Button3)
+	 checkwarp(newwire->points + newwire->number - 1);
    }
 }
 
