@@ -1,6 +1,6 @@
 /*-------------------------------------------------------------------------*/
 /* functions.c 								   */
-/* Copyright (c) 1998  Tim Edwards, Johns Hopkins University        	   */
+/* Copyright (c) 2000  Tim Edwards, Johns Hopkins University        	   */
 /*-------------------------------------------------------------------------*/
 
 /*-------------------------------------------------------------------------*/
@@ -31,16 +31,25 @@ extern Window  win;
 extern Pixmap STIPPLE[8];
 extern short eventmode;
 extern Clientdata areastruct;
+extern Globaldata xobjs;
 extern int *appcolors;
+extern short textpos, textend;
+extern objectpair *pushlist;
+extern short pushes;
 
 /*------------------------------------------------------------------------*/
 /* Declarations of functions defined externally to functions.c		  */
 /*------------------------------------------------------------------------*/
 
-extern short ULength();
+extern TextExtents ULength();
 extern void UTransformPoints();
 
 void UTransformbyCTM();
+void calcbboxvalues(objectptr, uchar, genericptr *);
+
+#ifdef SCHEMA
+void pinadjust (short, short *, short *, short);
+#endif
 
 /*-----------------------*/
 /* Arithmetic functions  */
@@ -400,27 +409,27 @@ short checkbounds()
    /* check window-to-user space */
 
    lval = 2 * (long)((float) (areastruct.width) / (*areastruct.vscale)) +
-	(long)areastruct.lowerleft->x;
+	(long)areastruct.pcorner->x;
    if (lval != (long)((short)lval)) return -1;
    lval = 2 * (long)((float) (areastruct.height) / (*areastruct.vscale)) +
-	(long)areastruct.lowerleft->y;
+	(long)areastruct.pcorner->y;
    if (lval != (long)((short)lval)) return -1;
 
    /* check user-to-window space */
 
-   lval = (long)((float)(objectdata->lowerleft.x - areastruct.lowerleft->x) *
+   lval = (long)((float)(objectdata->lowerleft.x - areastruct.pcorner->x) *
 	(*areastruct.vscale));
    if (lval != (long)((short)lval)) return -1;
    lval = (long)areastruct.height - (long)((float)(objectdata->lowerleft.y -
-	areastruct.lowerleft->y) * (*areastruct.vscale)); 
+	areastruct.pcorner->y) * (*areastruct.vscale)); 
    if (lval != (long)((short)lval)) return -1;
    UTransformbyCTM(DCTM, &(objectdata->lowerleft), &testpt, 1);
 
    lval = (long)((float)(objectdata->lowerleft.x + objectdata->width -
-	areastruct.lowerleft->x) * (*areastruct.vscale));
+	areastruct.pcorner->x) * (*areastruct.vscale));
    if (lval != (long)((short)lval)) return -1;
    lval = (long)areastruct.height - (long)((float)(objectdata->lowerleft.y +
-	objectdata->height - areastruct.lowerleft->y) * (*areastruct.vscale)); 
+	objectdata->height - areastruct.pcorner->y) * (*areastruct.vscale)); 
    if (lval != (long)((short)lval)) return -1;
 
    return 0;
@@ -434,9 +443,17 @@ void window_to_user(xw, yw, upt)
   short 	xw, yw;
   XPoint	*upt;
 {
-  upt->x = (short)((float) xw / (*areastruct.vscale)) + areastruct.lowerleft->x;
-  upt->y = (short)((float)(areastruct.height - yw) / (*areastruct.vscale)) + 
-	areastruct.lowerleft->y;
+  float tmpx, tmpy;
+
+  tmpx = (float)xw / (*areastruct.vscale) + (float)areastruct.pcorner->x;
+  tmpy = (float)(areastruct.height - yw) / (*areastruct.vscale) + 
+	(float)areastruct.pcorner->y;
+
+  tmpx += (tmpx > 0) ? 0.5 : -0.5;
+  tmpy += (tmpy > 0) ? 0.5 : -0.5;
+
+  upt->x = (short)tmpx;
+  upt->y = (short)tmpy;
 }
 
 /*------------------------------------------------------------------------*/
@@ -447,9 +464,17 @@ void user_to_window(upt, wpt)
   XPoint  upt;
   XPoint  *wpt;
 {
-  wpt->x = (short)((float)(upt.x - areastruct.lowerleft->x) * (*areastruct.vscale));
-  wpt->y = areastruct.height - (short)((float)(upt.y - areastruct.lowerleft->y)
-	* (*areastruct.vscale)); 
+  float tmpx, tmpy;
+
+  tmpx = (float)(upt.x - areastruct.pcorner->x) * (*areastruct.vscale);
+  tmpy = (float)areastruct.height - (float)(upt.y - areastruct.pcorner->y)
+	* (*areastruct.vscale); 
+
+  tmpx += (tmpx > 0) ? 0.5 : -0.5;
+  tmpy += (tmpy > 0) ? 0.5 : -0.5;
+
+  wpt->x = (short)tmpx;
+  wpt->y = (short)tmpy;
 }
 
 /*------------------------------------------------------------------------*/
@@ -478,17 +503,29 @@ short UTopTransScale(length)
 void u2u_snap(uvalue)
   XPoint	*uvalue;
 {
-   short tmpx, tmpy;
+   float tmpx, tmpy;
+   float tmpix, tmpiy;
 
    if (areastruct.snapto) {
-      tmpx = uvalue->x / areastruct.snapspace;
-      if (abs(uvalue->x) % areastruct.snapspace > areastruct.snapspace / 2) 
-	 tmpx += sign(uvalue->x);
-      tmpy = uvalue->y / areastruct.snapspace;
-      if (abs(uvalue->y) % areastruct.snapspace > areastruct.snapspace / 2) 
-	 tmpy += sign(uvalue->y);
-      uvalue->x = tmpx * areastruct.snapspace;
-      uvalue->y = tmpy * areastruct.snapspace;
+      tmpx = (float)uvalue->x / xobjs.pagelist[areastruct.page]->snapspace;
+      if (tmpx > 0)
+	 tmpix = (float)((int)(tmpx + 0.5));
+      else
+         tmpix = (float)((int)(tmpx - 0.5));
+
+      tmpy = (float)uvalue->y / xobjs.pagelist[areastruct.page]->snapspace;
+      if (tmpy > 0)
+         tmpiy = (float)((int)(tmpy + 0.5));
+      else
+         tmpiy = (float)((int)(tmpy - 0.5));
+
+      tmpix *= xobjs.pagelist[areastruct.page]->snapspace;
+      tmpix += (tmpix > 0) ? 0.5 : -0.5;
+      tmpiy *= xobjs.pagelist[areastruct.page]->snapspace;
+      tmpiy += (tmpiy > 0) ? 0.5 : -0.5;
+
+      uvalue->x = (int)tmpix;
+      uvalue->y = (int)tmpiy;
    }
 }
 
@@ -513,8 +550,7 @@ void snap(valuex, valuey, returnpt)
 /* Current transformation matrix manipulation routines			  */
 /*------------------------------------------------------------------------*/
 
-void UResetCTM(ctm)
-  Matrix *ctm;
+void UResetCTM(Matrix *ctm)
 {
    ctm->a = ctm->e = 1;
    ctm->b = ctm->d = 0;
@@ -523,6 +559,25 @@ void UResetCTM(ctm)
 
 /*------------------------------------------------------------------------*/
 
+void InvertCTM(Matrix *ctm)
+{
+   float det = ctm->a * ctm->e - ctm->b * ctm->d;
+   float tx = ctm->b * ctm->f - ctm->c * ctm->e;
+   float ty = ctm->d * ctm->c - ctm->a * ctm->f;
+
+   float tmpa = ctm->a;
+
+   ctm->b = -ctm->b / det;
+   ctm->d = -ctm->d / det;
+
+   ctm->a = ctm->e / det;
+   ctm->e = tmpa / det;
+   ctm->c = tx / det;
+   ctm->f = ty / det;
+}
+
+/*------------------------------------------------------------------------*/
+
 void UCopyCTM(fctm, tctm)
    Matrix *fctm, *tctm;
 {
@@ -544,11 +599,11 @@ void UMakeWCTM(ctm)
 {
    ctm->a *= (*areastruct.vscale);
    ctm->b *= (*areastruct.vscale);
-   ctm->c = (ctm->c - (float)areastruct.lowerleft->x) * (*areastruct.vscale);
+   ctm->c = (ctm->c - (float)areastruct.pcorner->x) * (*areastruct.vscale);
    
    ctm->d *= -(*areastruct.vscale);
    ctm->e *= -(*areastruct.vscale);
-   ctm->f = (float)areastruct.height + ((float)areastruct.lowerleft->y - ctm->f) *
+   ctm->f = (float)areastruct.height + ((float)areastruct.pcorner->y - ctm->f) *
 	    (*areastruct.vscale);
 }
 
@@ -601,6 +656,48 @@ void USlantCTM(ctm, beta)
 
 /*------------------------------------------------------------------------*/
 
+void UPreScaleCTM(ctm)
+  Matrix *ctm;
+{
+   if ((ctm->a < 0) || (ctm->a == 0 && ((ctm->b * ctm->d) < 0))) {
+      ctm->a = -ctm->a;
+      ctm->d = -ctm->d;
+   }
+   if (ctm->e > 0) {
+      ctm->e = -ctm->e;
+      ctm->b = -ctm->b;
+   }
+}
+
+/*------------------------------------------------------------------------*/
+/* Adjust justification and CTM as necessary for flip invariance	  */
+/*------------------------------------------------------------------------*/
+
+short flipadjust(short justify)
+{
+   short tmpjust = justify & (~FLIPINV);
+
+   if (justify & FLIPINV) {
+      if ((DCTM)->a < 0) {
+         if ((tmpjust & (RIGHT | NOTLEFT)) != NOTLEFT)
+            tmpjust ^= (RIGHT | NOTLEFT);
+      }   
+      else if ((DCTM)->a == 0) {
+	 if (((DCTM)->b * (DCTM)->d) < 0)
+            if ((tmpjust & (RIGHT | NOTLEFT)) != NOTLEFT)
+               tmpjust ^= (RIGHT | NOTLEFT);
+      }
+      if ((DCTM)->e > 0) {
+         if ((tmpjust & (TOP | NOTBOTTOM)) != NOTBOTTOM) 
+            tmpjust ^= (TOP | NOTBOTTOM);
+      }
+      UPreScaleCTM(DCTM);
+   }      
+   return tmpjust;
+}
+
+/*------------------------------------------------------------------------*/
+
 void UPreMultCTM(ctm, position, scale, rotate)
   Matrix *ctm;
   XPoint position;
@@ -698,6 +795,23 @@ void UTransformPoints(points, newpoints, number, atpt, scale, rotate)
    UTransformbyCTM(&LCTM, points, newpoints, number);
 }
 
+/*----------------------------------------------------*/
+/* Transform points inward to next hierarchical level */
+/*----------------------------------------------------*/
+
+void InvTransformPoints(points, newpoints, number, atpt, scale, rotate)
+  XPoint       *points, *newpoints, atpt;
+  short        number, rotate;
+  float        scale;
+{
+   Matrix LCTM;
+ 
+   UResetCTM(&LCTM);
+   UPreMultCTM(&LCTM, atpt, scale, rotate);
+   InvertCTM(&LCTM);
+   UTransformbyCTM(&LCTM, points, newpoints, number);
+}
+
 /*------------------------------------------------------------------------*/
 /* Translate wire coords to force wire to horizontal or vertical position */
 /*------------------------------------------------------------------------*/
@@ -715,9 +829,9 @@ void manhattanize(pospt, newwire)
    else pospt->y = tpoint->y;
 }
 
-/*-------------------------------------------------------------------------*/
-/* Bounding box calculation routines					   */
-/*-------------------------------------------------------------------------*/
+/*----------------------------------------------------------------------*/
+/* Bounding box calculation routines					*/
+/*----------------------------------------------------------------------*/
 
 void bboxcalc(testval, lowerval, upperval)
   short	testval, *lowerval, *upperval;
@@ -726,7 +840,9 @@ void bboxcalc(testval, lowerval, upperval)
    if (testval > *upperval) *upperval = testval;
 }
 
-/*------------------------------------------------------------------------*/
+/*----------------------------------------------------------------------*/
+/* Bounding box calculation for elements which can be part of a path	*/
+/*----------------------------------------------------------------------*/
 
 calcextents(bboxgen, llx, lly, urx, ury)
   genericptr *bboxgen;
@@ -767,111 +883,340 @@ calcextents(bboxgen, llx, lly, urx, ury)
 }
 
 /*-------------------------------------------------------------------------*/
-/* Bounding box calculation for an object				   */
+/* Bounding box calculation for a single generic element		   */
 /*-------------------------------------------------------------------------*/
 
-void calcbbox(localdata)
-  objectptr	localdata;
+void calcbboxsingle(genericptr *bboxgen, short *llx, short *lly, short *urx,
+		short *ury
+#ifdef SCHEMA
+		, short *pllx, short *plly, short *purx, short *pury
+#endif
+									)
 {
-   /* set starting bounds as maximum bounds of screen */
-
    XPoint points[4], npoints[4];
    short i, j;
-   genericptr *bboxgen;
 
-   short llx, lly, urx, ury;
+   /* For each screen element, compute the extents and revise bounding	*/
+   /* box points, if necessary. 					*/
+
+   switch((*bboxgen)->type) {
+
+      case(OBJECT):
 
 #ifdef SCHEMA
-   short pllx, plly, purx, pury;  /* pin label extents */
-
-   pllx = plly = 32767;
-   purx = pury = -32768;
-#endif
-
-   llx = lly = 32767;
-   urx = ury = -32768;
-
-   /* now, for each screen element, compute the extents and revise bounding
-      box points, if necessary. */
-
-   for (bboxgen = localdata->plist; bboxgen < localdata->plist +
-		localdata->parts; bboxgen++) {
-      switch((*bboxgen)->type) {
-
-	 case(OBJECT):
-            points[0].x = points[1].x = TOOBJINST(bboxgen)->thisobject->lowerleft.x;
-            points[1].y = points[2].y = TOOBJINST(bboxgen)->thisobject->lowerleft.y
+         points[0].x = points[1].x = TOOBJINST(bboxgen)->thisobject->lleft2.x;
+         points[1].y = points[2].y = TOOBJINST(bboxgen)->thisobject->lleft2.y
+		+ TOOBJINST(bboxgen)->thisobject->height2;
+         points[2].x = points[3].x = TOOBJINST(bboxgen)->thisobject->lleft2.x
+		+ TOOBJINST(bboxgen)->thisobject->width2;
+         points[0].y = points[3].y = TOOBJINST(bboxgen)->thisobject->lleft2.y;
+#else
+         points[0].x = points[1].x = TOOBJINST(bboxgen)->thisobject->lowerleft.x;
+         points[1].y = points[2].y = TOOBJINST(bboxgen)->thisobject->lowerleft.y
 		+ TOOBJINST(bboxgen)->thisobject->height;
-            points[2].x = points[3].x = TOOBJINST(bboxgen)->thisobject->lowerleft.x
+         points[2].x = points[3].x = TOOBJINST(bboxgen)->thisobject->lowerleft.x
 		+ TOOBJINST(bboxgen)->thisobject->width;
-            points[0].y = points[3].y = TOOBJINST(bboxgen)->thisobject->lowerleft.y;
-            UTransformPoints(points, npoints, 4, TOOBJINST(bboxgen)->position,
+         points[0].y = points[3].y = TOOBJINST(bboxgen)->thisobject->lowerleft.y;
+#endif
+         UTransformPoints(points, npoints, 4, TOOBJINST(bboxgen)->position,
 		 TOOBJINST(bboxgen)->scale, TOOBJINST(bboxgen)->rotation); 
-            for (j = 0; j < 4; j++) {
-               bboxcalc(npoints[j].x, &llx, &urx);
-               bboxcalc(npoints[j].y, &lly, &ury);
-            }
-	    break;
+         for (j = 0; j < 4; j++) {
+            bboxcalc(npoints[j].x, llx, urx);
+            bboxcalc(npoints[j].y, lly, ury);
+         }
+	 break;
 
-	 case(LABEL): {
-            float tmpwidth = ULength(TOLABEL(bboxgen)->string, 0.0, 0);
+      case(LABEL):
+	 {
+	    TextExtents tmpext;
+            tmpext = ULength(TOLABEL(bboxgen)->string, 0.0, 0, NULL);
             points[0].x = points[1].x = (TOLABEL(bboxgen)->justify & NOTLEFT ? 
-	       (TOLABEL(bboxgen)->justify & RIGHT ? -tmpwidth : -tmpwidth / 2) : 0);
+	       (TOLABEL(bboxgen)->justify & RIGHT ? -tmpext.width :
+		-tmpext.width / 2) : 0);
             points[0].y = points[3].y = (TOLABEL(bboxgen)->justify & NOTBOTTOM ?
-	       (TOLABEL(bboxgen)->justify & TOP ? -TEXTHEIGHT : -HALFHEIGHT) : 0);
-            points[1].y = points[2].y = points[0].y + TEXTHEIGHT;
-            points[2].x = points[3].x = points[0].x + tmpwidth;
+	       (TOLABEL(bboxgen)->justify & TOP ? -tmpext.ascent :
+		-(tmpext.ascent + tmpext.base) / 2) : -tmpext.base);
+            points[1].y = points[2].y = points[0].y + tmpext.ascent - tmpext.descent;
+            points[2].x = points[3].x = points[0].x + tmpext.width;
 
             UTransformPoints(points, npoints, 4, TOLABEL(bboxgen)->position,
 		TOLABEL(bboxgen)->scale, TOLABEL(bboxgen)->rotation); 
 #ifdef SCHEMA
-	    /* separate bounding box for pinlabels */
+	    /* separate bounding box for pinlabels and infolabels */
 
-	    if (TOLABEL(bboxgen)->pin == True) {
+	    if (TOLABEL(bboxgen)->pin) {
                for (j = 0; j < 4; j++) {
-                  bboxcalc(npoints[j].x, &pllx, &purx);
-                  bboxcalc(npoints[j].y, &plly, &pury);
+		  pinadjust(TOLABEL(bboxgen)->justify, &npoints[j].x,
+			&npoints[j].y, 1);
+                  bboxcalc(npoints[j].x, pllx, purx);
+                  bboxcalc(npoints[j].y, plly, pury);
                }
 	    }
-	    else {
+	    else
 #endif
             for (j = 0; j < 4; j++) {
-               bboxcalc(npoints[j].x, &llx, &urx);
-               bboxcalc(npoints[j].y, &lly, &ury);
+               bboxcalc(npoints[j].x, llx, urx);
+               bboxcalc(npoints[j].y, lly, ury);
             }
-#ifdef SCHEMA
-	    }
-#endif
-            } break;
+         } break;
 
-	 case(PATH): {
-	    genericptr *pathc;
-	    for (pathc = TOPATH(bboxgen)->plist; pathc < TOPATH(bboxgen)->plist
+      case(PATH): {
+	 genericptr *pathc;
+	 for (pathc = TOPATH(bboxgen)->plist; pathc < TOPATH(bboxgen)->plist
 		  + TOPATH(bboxgen)->parts; pathc++)
-	       calcextents(pathc, &llx, &lly, &urx, &ury);
-	    } break;
+	       calcextents(pathc, llx, lly, urx, ury);
+	 } break;
 
-	 default:
-	    calcextents(bboxgen, &llx, &lly, &urx, &ury);
+      default:
+	 calcextents(bboxgen, llx, lly, urx, ury);
+   }
+}
+
+/*------------------------------------------------------*/
+/* Find if an object is in the specified library	*/
+/*------------------------------------------------------*/
+
+Boolean object_in_library(short libnum, objectptr thisobject)
+{
+   short i;
+   objectptr lobj;
+
+   for (i = 0; i < xobjs.userlibs[libnum].number; i++) {
+      if (*(xobjs.userlibs[libnum].library + i) == thisobject)
+	 return True;
+   }
+   return False;
+}
+
+/*-----------------------------------------------------------*/
+/* Find if an object is in the hierarchy of the given object */
+/* Returns the number (position in plist) or -1 if not found */
+/*-----------------------------------------------------------*/
+
+short find_object(objectptr pageobj, objectptr thisobject)
+{
+   short i, j;
+   genericptr *pelem;
+
+   for (i = 0; i < pageobj->parts; i++) {
+      pelem = pageobj->plist + i;
+      if ((*pelem)->type == OBJECT) {
+	 if ((TOOBJINST(pelem))->thisobject == thisobject)
+	    return i;
+	 else if ((j = find_object((TOOBJINST(pelem))->thisobject, thisobject)) >= 0)
+	    return i;  /* was j---is this the right fix? */
       }
    }
-   
-   /* default action if there are no elements */
+   return -1;
+}
 
-   if (localdata->parts == 0) {
-      llx = areastruct.lowerleft->x;
-      lly = areastruct.lowerleft->y;
-      urx = llx + areastruct.width / (*areastruct.vscale);
-      ury = lly + areastruct.height / (*areastruct.vscale);
+/*------------------------------------------------------*/
+/* Find all pages and libraries containing this object	*/
+/* and update accordingly.  If this object is a page,	*/
+/* just update the page directory.			*/
+/*------------------------------------------------------*/
+
+void updatepagebounds(objectptr thisobject)
+{
+   short i, j;
+
+   if ((i = is_page(thisobject)) >= 0) {
+      if (xobjs.pagelist[i]->background.name != (char *)NULL)
+         backgroundbbox(i);
+      updatepagelib(PAGELIB, i);
+   }
+   else {
+      for (i = 0; i < xobjs.pages; i++)
+         if (xobjs.pagelist[i]->pageobj != NULL)
+            if ((j = find_object(xobjs.pagelist[i]->pageobj, thisobject)) >= 0) {
+	       calcbboxvalues(xobjs.pagelist[i]->pageobj, (uchar)0,
+		       (genericptr *)(xobjs.pagelist[i]->pageobj->plist + j));
+	       updatepagelib(PAGELIB, i);
+	    }
+
+      for (i = 0; i < xobjs.numlibs; i++)
+         if (object_in_library(i, thisobject))
+	    composelib(i + LIBRARY);
+   }
+}
+
+/*--------------------------------------------------------------*/
+/* Calculate bbox on all elements of the given object		*/
+/*--------------------------------------------------------------*/
+
+void calcbbox(objectptr bobj)
+{
+   calcbboxvalues(bobj, (uchar)0, (genericptr *)NULL);
+   if (bobj == objectdata)
+      updatepagebounds(objectdata);
+}
+
+/*--------------------------------------------------------------*/
+/* Calculate bbox for an object with a parameter which was just	*/
+/* changed.  If the parameter was a single-instance		*/
+/* substitution, only the page should be updated.  If the	*/
+/* parameter was a default value, the library should be updated */
+/* and any pages containing the object where the parameter	*/
+/* takes the default value.					*/
+/*--------------------------------------------------------------*/
+
+void calcbboxparam(objectptr bobj, int mode)
+{
+   short i, j;
+
+   calcbboxvalues(bobj, (uchar)0, (genericptr *)NULL);
+
+   /* this is a modified version of updatepagebounds() 	*/
+   /* mode == 0:  change bounds on pagelib and the page	*/
+   /* containing this object *instance*			*/
+
+   if (mode == 0) {
+      /* pick up calling object and instance from the edit stack */
+      objectptr pageobj = (pushlist + pushes - 1)->thisobject;
+      objinstptr pinst = (pushlist + pushes - 1)->thisinst;
+
+      for (j = 0; j < pageobj->parts; j++)
+	 if (*(pageobj->plist + j) == (genericptr)pinst) break;
+      if (j == pageobj->parts) {
+	 fprintf(stderr, "Error: Calling page does not contain expected objinst!\n");
+	 return;
+      }
+      calcbboxvalues(pageobj, (uchar)0, (genericptr *)(pageobj->plist + j));
+      for (i = 0; i < xobjs.pages; i++)
+	 if (xobjs.pagelist[i]->pageobj == pageobj) {
+            updatepagelib(PAGELIB, i);
+	    break;
+	 }
+      if (i == xobjs.pages)
+	 fprintf(stderr, "Error: Calling page not found!\n");
+   }
+
+   /* mode == 1:  change bounds on pagelib and all pages	*/
+   /*   containing this *object* if and only if the object	*/
+   /*   instance takes the default value.  Also update the	*/
+   /*   library page.						*/
+
+   /* It actually appears to be impossible to satisfy all 	*/
+   /* possible conditions and ultimately, the bounding box is	*/
+   /* going to be wrong sometimes.				*/
+
+   else {
+      for (i = 0; i < xobjs.pages; i++)
+         if (xobjs.pagelist[i]->pageobj != NULL)
+            if ((j = find_object(xobjs.pagelist[i]->pageobj, objectdata)) >= 0) {
+	       objinstptr cinst = TOOBJINST(xobjs.pagelist[i]->pageobj->plist + j);
+	       /* Really, we'd like to recalculate the bounding box only if the */
+	       /* parameter value is the default value which was just changed.	*/
+	       /* However, then any non-default values may contain the wrong	*/
+	       /* substitutions.						*/
+	       if (cinst->num_params == 0) {
+	          calcbboxvalues(xobjs.pagelist[i]->pageobj, (uchar)0,
+		       (genericptr *)(xobjs.pagelist[i]->pageobj->plist + j));
+	          updatepagelib(PAGELIB, i);
+	       }
+	    }
+
+      for (i = 0; i < xobjs.numlibs; i++)
+         if (object_in_library(i, objectdata))
+	    composelib(i + LIBRARY);
+   }
+}
+
+/*--------------------------------------------------------------*/
+/* Calculate bbox on the given element of the specified object	*/
+/* (assumes that given element can only *increase* the bbox)	*/
+/*--------------------------------------------------------------*/
+
+void singlebbox(objectptr localdata, genericptr *gelem)
+{
+   calcbboxvalues(localdata, (uchar)0, (genericptr *)gelem);
+   updatepagebounds(localdata);
+}
+
+/*----------------------------------------------------------------------*/
+/* Extend bounding box based on selected elements only			*/
+/*----------------------------------------------------------------------*/
+
+void calcbboxselect()
+{
+   short *bsel;
+   for (bsel = areastruct.selectlist; bsel < areastruct.selectlist +
+	areastruct.selects; bsel++) {
+      singlebbox(objectdata, objectdata->plist + *bsel);
+   }
+   updatepagebounds(objectdata);
+}
+
+/*--------------------------------------------------------------*/
+/* Update Bounding box on an object or single element		*/
+/*--------------------------------------------------------------*/
+
+void calcbboxvalues(objectptr localdata, uchar mode, genericptr *newelement)
+{
+   genericptr *bboxgen;
+   short llx, lly, urx, ury;
+#ifdef SCHEMA
+   short pllx, plly, purx, pury;  /* pin label extents */
+#endif
+
+   /* no action if there are no elements */
+   if (localdata->parts == 0) return;
+
+   if (newelement == (genericptr *)NULL) {
+
+      /* set starting bounds as maximum bounds of screen */
+
+#ifdef SCHEMA
+      pllx = plly = 32767;
+      purx = pury = -32768;
+#endif
+
+      llx = lly = 32767;
+      urx = ury = -32768;
+
+      for (bboxgen = localdata->plist; bboxgen < localdata->plist +
+		localdata->parts; bboxgen++)
+         calcbboxsingle(bboxgen, &llx, &lly, &urx, &ury
+#ifdef SCHEMA
+						, &pllx, &plly, &purx, &pury
+#endif
+										);
+   }
+   else {
+
+#ifdef SCHEMA
+      llx = localdata->lleft2.x;
+      lly = localdata->lleft2.y;
+      urx = localdata->width2 + llx;
+      ury = localdata->height2 + lly;
+
+      pllx = localdata->lowerleft.x;
+      plly = localdata->lowerleft.y;
+      purx = localdata->width + pllx;
+      pury = localdata->height + plly;
+#else
+      llx = localdata->lowerleft.x;
+      lly = localdata->lowerleft.y;
+      urx = localdata->width + llx;
+      ury = localdata->height + lly;
+#endif
+
+      calcbboxsingle(newelement, &llx, &lly, &urx, &ury
+#ifdef SCHEMA
+						, &pllx, &plly, &purx, &pury
+#endif
+										);
    }
 
 #ifdef SCHEMA
    /* bounding box without pin info. . */
 
-   localdata->lleft2.x = llx;
-   localdata->lleft2.y = lly;
-   localdata->width2 = urx - llx;
-   localdata->height2 = ury - lly;
+   if (!mode) {
+         localdata->lleft2.x = llx;
+         localdata->lleft2.y = lly;
+         localdata->width2 = urx - llx;
+         localdata->height2 = ury - lly;
+   }
 
    /* . . . and with . . . */
 
@@ -887,63 +1232,64 @@ void calcbbox(localdata)
    localdata->lowerleft.y = lly;
    localdata->width = urx - llx;
    localdata->height = ury - lly;
-
-   /* avoid width or height of zero---use of TEXTHEIGHT is arbitrary. . . */
-
-   if (localdata->width == 0) {
-      localdata->width = TEXTHEIGHT;
-      localdata->lowerleft.x -= HALFHEIGHT;
-   }
-   if (localdata->height == 0) {
-      localdata->height = TEXTHEIGHT;
-      localdata->lowerleft.y -= HALFHEIGHT;
-   }
 }
 
-/*--------------------------------------------------------*/
-/* Modify areastruct on basis of bounding box calculation */
-/*--------------------------------------------------------*/
+/*------------------------------------------------------*/
+/* Center an object in the viewing window		*/
+/*------------------------------------------------------*/
 
-void zoomview(w, mode, nulldata)
-  Widget 	 w;
-  unsigned short mode;
-  caddr_t	 nulldata;
+void centerview(objectptr localdata)
 {
-   calcbbox(objectdata);
+   float fitwidth = (float)areastruct.width / ((float)localdata->width + 
+	       2 * DEFAULTGRIDSPACE);
+   float fitheight = (float)areastruct.height / ((float)localdata->height +
+	       2 * DEFAULTGRIDSPACE);
 
-   if (eventmode == NORMAL_MODE || eventmode == POPUP_MODE ||
-	 eventmode == COPY2_MODE || eventmode == PUSH_MODE ||
-	 eventmode == PRESS_MODE || eventmode == CATALOG_MODE) {
+   localdata->viewscale = (fitwidth < fitheight) ?
+		 min(MINAUTOSCALE, fitwidth) : min(MINAUTOSCALE, fitheight);
 
-      if ((mode == 1) || (areastruct.vscale !=
-	       &(areastruct.viewscale[areastruct.page]))) {
-         float fitwidth = (float)areastruct.width / (float)(objectdata->width + 
-	       2 * areastruct.gridspace);
-         float fitheight = (float)areastruct.height / (float)(objectdata->height +
-	       2 * areastruct.gridspace);
+   localdata->pcorner.x = localdata->lowerleft.x - (areastruct.width
+	       / localdata->viewscale - localdata->width) / 2;
+   localdata->pcorner.y = localdata->lowerleft.y - (areastruct.height
+	       / localdata->viewscale - localdata->height) / 2;
+}
 
-         if (fitwidth < fitheight) 
-            (*areastruct.vscale) = min(MINAUTOSCALE, fitwidth);
-         else 
-            (*areastruct.vscale) = min(MINAUTOSCALE, fitheight);
+/*-----------------------------------------------------------*/
+/* Refresh the window and scrollbars and write the page name */
+/*-----------------------------------------------------------*/
 
-         areastruct.lowerleft->x = objectdata->lowerleft.x - (areastruct.width
-	       / (*areastruct.vscale) - objectdata->width) / 2;
-         areastruct.lowerleft->y = objectdata->lowerleft.y - (areastruct.height
-	       / (*areastruct.vscale) - objectdata->height) / 2;
-      }
-      UResetCTM(DCTM);
-      UMakeWCTM(DCTM);
+XtCallbackProc refresh(Widget bw, caddr_t clientdata, caddr_t calldata)
+{
+   drawarea(NULL, objectdata, NULL);
+   drawhbar(areastruct.scrollbarh, NULL, NULL);
+   drawvbar(areastruct.scrollbarv, NULL, NULL);
+   printname(objectdata);
+}
 
-      drawarea(w, NULL, NULL);
-      drawhbar(areastruct.scrollbarh, NULL, Number(1));
-      drawvbar(areastruct.scrollbarv, NULL, Number(1));
+/*------------------------------------------------------*/
+/* Center the current page in the viewing window	*/
+/*------------------------------------------------------*/
+
+void zoomview(w, clientdata, calldata)
+  Widget 	 w;
+  caddr_t	 clientdata, calldata;
+{
+   if (eventmode == NORMAL_MODE || eventmode == COPY2_MODE ||
+	 eventmode == PUSH_MODE || eventmode == PRESS_MODE ||
+	 eventmode == CATALOG_MODE || eventmode == FONTCAT_MODE ||
+	 eventmode == FONTCAT2_MODE) {
+
+      centerview(objectdata);
+      newmatrix();
+      areastruct.lastbackground = NULL;
+      renderbackground();
+      refresh(NULL, NULL, NULL);
    }
 }
 
-/*-------------------------------------------------------------------------*/
-/* Basic X Graphics Routines in the User coordinate system		   */
-/*-------------------------------------------------------------------------*/
+/*---------------------------------------------------------*/
+/* Basic X Graphics Routines in the User coordinate system */
+/*---------------------------------------------------------*/
 
 void UDrawSimpleLine(localdata, pt1, pt2)
   objinstptr	localdata;
@@ -963,7 +1309,7 @@ void UDrawLine(localdata, pt1, pt2)
   objinstptr	localdata;
   XPoint	*pt1, *pt2;
 {
-   short tmpwidth = UTopTransScale(areastruct.wirewidth[areastruct.page]);
+   short tmpwidth = UTopTransScale(xobjs.pagelist[areastruct.page]->wirewidth);
 
    XSetLineAttributes(dpy, areastruct.gc, tmpwidth >= 2.0 ? (int)tmpwidth : 
 	0, LineSolid, CapRound, JoinBevel);
@@ -986,41 +1332,94 @@ void UDrawX(curlabel)
 }
 
 /*-------------------------------------------------------------------------*/
-/* Draw line for editing text (position of cursor in string is given by    */
-/*   areastruct.origin.(x,y)						   */
+/* Adjust a pinlabel position to account for pad spacing		   */
 /*-------------------------------------------------------------------------*/
 
-void UDrawTLine(curlabel)
+#ifdef SCHEMA
+
+void pinadjust (short justify, short *xpoint, short *ypoint, short dir)
+{
+   int delx, dely;
+
+   dely = (justify & NOTBOTTOM) ?
+            ((justify & TOP) ? -PADSPACE : 0) : PADSPACE;
+   delx = (justify & NOTLEFT) ?
+            ((justify & RIGHT) ? -PADSPACE : 0) : PADSPACE;
+
+   if (xpoint != NULL) *xpoint += (dir > 0) ? delx : -delx;
+   if (ypoint != NULL) *ypoint += (dir > 0) ? dely : -dely;
+}
+
+#endif
+
+/*----------------------------------------------------------------------*/
+/* Draw line for editing text (position of cursor in string is given by */
+/*   tpos (2nd parameter)						*/
+/*----------------------------------------------------------------------*/
+
+void UDrawTextLine(curlabel, tpos)
    labelptr	curlabel;
+   short	tpos;
 {
    XPoint  upt, points[2]; /* top and bottom of text cursor line */
-   short   xdist, tmplength;
+   short   xdist, xbase, tmpjust;
+   TextExtents tmpext;
+
+   /* correct for position, rotation, scale, and flip invariance of text */
+
+   UPushCTM();
+   UPreMultCTM(DCTM, curlabel->position, curlabel->scale, curlabel->rotation);
+   tmpjust = flipadjust(curlabel->justify);
 
    XSetFunction(dpy, areastruct.gc, GXxor);
    XSetForeground(dpy, areastruct.gc, AUXCOLOR ^ BACKGROUND);
 
-   tmplength = ULength(curlabel->string, 0.0, 0);
-   xdist = ULength(curlabel->string, 0.0, 1);
+   tmpext = ULength(curlabel->string, 0.0, tpos, NULL);
+   xdist = tmpext.width;
+   xbase = tmpext.base;
+   tmpext = ULength(curlabel->string, 0.0, 0, NULL);
 
-   points[0].x = (curlabel->justify & NOTLEFT ?
-        (curlabel->justify & RIGHT ? -tmplength : -tmplength >> 1) : 0)
+   points[0].x = (tmpjust & NOTLEFT ?
+        (tmpjust & RIGHT ? -tmpext.width : -tmpext.width >> 1) : 0)
 	+ xdist;
-   points[0].y = (curlabel->justify & NOTBOTTOM ?
-        (curlabel->justify & TOP ? -TEXTHEIGHT : -HALFHEIGHT) : 0) - 3;
+   points[0].y = (tmpjust & NOTBOTTOM ?
+        (tmpjust & TOP ? -tmpext.ascent : -(tmpext.ascent + tmpext.base) / 2)
+	: -tmpext.base) + xbase - 3;
    points[1].x = points[0].x;
    points[1].y = points[0].y + TEXTHEIGHT + 6;
 
-   /* correct for position, rotation and scale of text */
+#ifdef SCHEMA
+   if (curlabel->pin) {
+      pinadjust(tmpjust, &(points[0].x), &(points[0].y), 1);
+      pinadjust(tmpjust, &(points[1].x), &(points[1].y), 1);
+   }
+#endif
+
+   /* draw the line */
 
-   UPushCTM();
-   UPreMultCTM(DCTM, curlabel->position, curlabel->scale, curlabel->rotation);
    UDrawLine(areastruct.topobject, &points[0], &points[1]);
    UPopCTM();
 
    UDrawX(curlabel);
 }
 
-/*-------------------------------------------------------------------------*/
+/*-----------------------------------------------------------------*/
+/* Draw lines for editing text when multiple characters are chosen */
+/*-----------------------------------------------------------------*/
+
+void UDrawTLine(curlabel)
+   labelptr	curlabel;
+{
+   UDrawTextLine(curlabel, textpos);
+   if ((textend > 0) && (textend < textpos)) {
+      UDrawTextLine(curlabel, textend);
+   }
+}
+
+/*----------------------*/
+/* Draw an X		*/
+/*----------------------*/
+
 void UDrawXLine(opt, cpt)
   XPoint	opt, cpt;
 {
@@ -1043,17 +1442,6 @@ void UDrawXLine(opt, cpt)
    XSetForeground(dpy, areastruct.gc, areastruct.gccolor);
 }
 
-/*-------------------------------------------------------------------------*/
-void UDrawLines(localdata, points, number)
-  objinstptr	localdata;
-  XPoint	*points;
-  short		number;
-{
-   XPoint newpoints[MAXSEGS];
-   UTransformbyCTM(DCTM, points, newpoints, number);
-   XDrawLines(dpy, win, areastruct.gc, newpoints, number, CoordModeOrigin);
-}
-
 /*-------------------------------------------------------------------------*/
 void UDrawBox(origin, corner)
   XPoint	origin, corner;
@@ -1073,6 +1461,25 @@ void UDrawBox(origin, corner)
    XDrawLine(dpy, win, areastruct.gc, wcorn.x, worig.y, worig.x, worig.y);
 }
 
+/*-------------------------------------------------------------------------*/
+void UDrawBBox(objectptr bbobj)
+{
+   XPoint	origin = bbobj->lowerleft;
+   XPoint	worig, wcorn, corner;
+
+   corner.x = origin.x + bbobj->width;
+   corner.y = origin.y + bbobj->height;
+
+   user_to_window(origin, &worig);
+   user_to_window(corner, &wcorn);
+
+   XSetForeground(dpy, areastruct.gc, BBOXCOLOR);
+   XDrawLine(dpy, win, areastruct.gc, worig.x, worig.y, worig.x, wcorn.y);
+   XDrawLine(dpy, win, areastruct.gc, worig.x, wcorn.y, wcorn.x, wcorn.y);
+   XDrawLine(dpy, win, areastruct.gc, wcorn.x, wcorn.y, wcorn.x, worig.y);
+   XDrawLine(dpy, win, areastruct.gc, wcorn.x, worig.y, worig.x, worig.y);
+}
+
 /*-------------------------------------------------------------------------*/
 /* Fill and/or draw a border around the stroking path			   */
 /*-------------------------------------------------------------------------*/
@@ -1087,7 +1494,7 @@ void strokepath(pathlist, number, style, width)
    char         dashstring[3];
    short        tmpwidth, i;
 
-   tmpwidth = max(1, UTopTransScale(areastruct.wirewidth[areastruct.page] *
+   tmpwidth = max(1, UTopTransScale(xobjs.pagelist[areastruct.page]->wirewidth *
 	width));
 
    if (style & FILLED || (!(style & FILLED) && style & OPAQUE)) {
@@ -1236,7 +1643,9 @@ void UDrawPath(localdata, thepath)
    free(tmppoints);
 }
 
-/*-------------------------------------------------------------------------*/
+/*----------------------------------------------------------------------*/
+/* Main recursive object drawing routine				*/
+/*----------------------------------------------------------------------*/
 
 void UDrawObject(localdata, theobject, level, passcolor)
   objinstptr	localdata;
@@ -1252,6 +1661,9 @@ void UDrawObject(localdata, theobject, level, passcolor)
    XPoint 	bboxin, bboxout[2];
    Boolean	xm, ym;
 
+   /* replace parameters (if any) from the instance into the object */
+   if (level != 0) psubstitute(localdata);
+
    /* add self to hierarchy list */
 
    areastruct.hlevels++;
@@ -1278,28 +1690,27 @@ void UDrawObject(localdata, theobject, level, passcolor)
 
      /* draw all of the elements */
    
-     tmpwidth = UTopTransScale(areastruct.wirewidth[areastruct.page]);
+     tmpwidth = UTopTransScale(xobjs.pagelist[areastruct.page]->wirewidth);
      XSetLineAttributes(dpy, areastruct.gc, tmpwidth >= 2.0 ? (int)tmpwidth :
 		0, LineSolid, CapRound, JoinBevel);
 
      for (areagen = theobject->plist; areagen < theobject->plist +
            theobject->parts; areagen++) {
 
-       if (defaultcolor != DOFORALL)
+       if (defaultcolor != DOFORALL) {
 	  if ((*areagen)->color != curcolor) {
-	     if ((*areagen)->color == DEFAULTCOLOR) {
+	     if ((*areagen)->color == DEFAULTCOLOR)
 		curcolor = defaultcolor;
-	        XSetForeground(dpy, areastruct.gc, defaultcolor);
-	     }
-	     else {
+	     else
 		curcolor = (*areagen)->color;
-	        XSetForeground(dpy, areastruct.gc, curcolor);
-	     }
+	     XTopSetForeground(curcolor);
 	  }
+       }
 
        switch((*areagen)->type) {
 	  case(POLYGON):
-             UDrawPolygon(localdata, TOPOLY(areagen));
+	     if (level == 0 || !((TOPOLY(areagen))->style & BBOX))
+                UDrawPolygon(localdata, TOPOLY(areagen));
 	     break;
    
 	  case(SPLINE):
@@ -1321,10 +1732,9 @@ void UDrawObject(localdata, theobject, level, passcolor)
    
   	  case(LABEL): 
 #ifdef SCHEMA
-	     if (areastruct.schemon == False || level == 0 ||
-		    TOLABEL(areagen)->pin == False)
+	     if (level == 0 || TOLABEL(areagen)->pin == False)
 #endif
-             UDrawString(localdata, TOLABEL(areagen));
+             UDrawString(localdata, TOLABEL(areagen), curcolor);
 	     break;
        }
      }
