@@ -1,6 +1,6 @@
 /*-------------------------------------------------------------------------*/
 /* selection.c --- xcircuit routines handling element selection etc.	   */
-/* Copyright (c) 1998  Tim Edwards, Johns Hopkins University       	   */
+/* Copyright (c) 2000  Tim Edwards, Johns Hopkins University       	   */
 /*-------------------------------------------------------------------------*/
 
 /*-------------------------------------------------------------------------*/
@@ -36,17 +36,37 @@ extern Cursor	appcursors[NUM_CURSORS];
 extern Clientdata areastruct;
 extern char _STR[150];
 extern short attachto;
-extern long finddist();
+extern short textpos, textend;
 
 #define RADFAC 0.0174532925199   /* pi / 180 */
 
-/*------------------------------------------------------------------------*/
-/* Declarations of functions defined in selection.c	 	  	  */
-/*------------------------------------------------------------------------*/
+/*----------------------------------------------------------------------*/
+/* Declarations of functions 						*/
+/*----------------------------------------------------------------------*/
 
-short *objectselect();
+extern long finddist();
+extern TextExtents ULength();
+
+short *selectobject(short, objectptr);
 XtEventHandler trackselarea();
 
+/*----------------------------------------------------------------------*/
+/* Change filter to determine what types can be selected		*/
+/*----------------------------------------------------------------------*/
+
+XtCallbackProc selectfilter(Widget w, caddr_t value, caddr_t calldata)
+{
+   short bitwise = (int) value & 0xffff;
+   Boolean bval = (areastruct.filter & bitwise) ? True : False;
+
+   if (bval)
+      areastruct.filter &= ~bitwise;
+   else
+      areastruct.filter |= bitwise;
+
+   toggle(w, &bval, calldata);
+}
+
 /*----------------------------------------------------------------------------*/
 /* Look at select stack to see if there are any selects; call select, if not. */
 /*----------------------------------------------------------------------------*/
@@ -55,6 +75,8 @@ Boolean checkselect(value)
   short value;
 {
    short *check;
+ 
+   value &= areastruct.filter;	/* apply the selection filter */
 
    if (areastruct.selects == 0) objectselect(value);
    if (areastruct.selects == 0) return False;
@@ -69,20 +91,11 @@ Boolean checkselect(value)
 /* Draw a selected item */
 /*----------------------*/
 
-void easydraw(instance, mode) 
-  short instance;
-  short mode;
+void geneasydraw(short instance, short mode, objectptr curobj)
 {
-   genericptr elementptr = *(objectdata->plist + instance);
+   genericptr elementptr = *(curobj->plist + instance);
 
-   switch (SELECTTYPE(&instance)) {
-      case OBJECT:
-         UDrawObject((objinstptr)elementptr, 
-	    ((objinstptr)elementptr)->thisobject, SINGLE, mode);
-         break;
-      case LABEL:
-         UDrawString(areastruct.topobject, (labelptr)elementptr);
-	 break;
+   switch ((*(curobj->plist + instance))->type) {
       case ARC:
          UDrawArc(areastruct.topobject, (arcptr)elementptr);
 	 break;
@@ -94,6 +107,14 @@ void easydraw(instance, mode)
 	 break;
       case PATH:
 	 UDrawPath(areastruct.topobject, (pathptr)elementptr);
+	 break;
+      case LABEL:
+         UDrawString(areastruct.topobject, (labelptr)elementptr, mode);
+	 break;
+      case OBJECT:
+         UDrawObject((objinstptr)elementptr, 
+	    ((objinstptr)elementptr)->thisobject, SINGLE, mode);
+         break;
    }
 }
 
@@ -101,15 +122,14 @@ void easydraw(instance, mode)
 /* Draw a selected item, including selection color */
 /*-------------------------------------------------*/
 
-void drawselected(newselect)
-  short *newselect;
+void gendrawselected(short *newselect, objectptr curobj)
 {
    XSetFunction(dpy, areastruct.gc, GXcopy);
    XSetForeground(dpy, areastruct.gc, BACKGROUND);
-   easydraw(*newselect, DOFORALL);
+   geneasydraw(*newselect, DOFORALL, curobj);
    XSetFunction(dpy, areastruct.gc, GXxor);
    XSetForeground(dpy, areastruct.gc, SELECTCOLOR ^ BACKGROUND);
-   easydraw(*newselect, DOFORALL);
+   geneasydraw(*newselect, DOFORALL, curobj);
 }
 
 /*---------------------------------------------------*/
@@ -173,6 +193,25 @@ setoptionmenu()
    }
 }
 
+/*-------------------------------------*/
+/* Test of point being inside of a box */
+/*-------------------------------------*/
+
+int test_insideness(int tx, int ty, XPoint *boxpoints) {
+   int i, stval = 0; 
+   XPoint *pt1, *pt2;
+   int stdir;   
+
+   for (i = 0; i < 4; i++) {
+      pt1 = boxpoints + i;
+      pt2 = boxpoints + ((i + 1) % 4);
+      stdir = (pt2->x - pt1->x) * (ty - pt1->y)
+		- (pt2->y - pt1->y) * (tx - pt1->x);
+      stval += sign(stdir);
+   }
+   return (abs(stval) == 4) ? 1 : 0;
+}
+
 /*--------------------------------------------*/
 /* Search for selection among path components */
 /*--------------------------------------------*/
@@ -199,6 +238,8 @@ Boolean pathselect(curgen, class)
    long		newdist;
    Boolean	selected = False;
 
+   class &= areastruct.filter;	/* apply the selection filter */
+
    if ((*curgen)->type == (class & ARC)) {
 
       /* look among the arcs */
@@ -280,8 +321,7 @@ Boolean pathselect(curgen, class)
 /* Select one of the elements on-screen */
 /*--------------------------------------*/
 
-short *objectselect(class)
-  short 	class;
+short *selectobject(short class, objectptr selobj)
 {
    short	*newselect, *foundtmp, *foundlist;
 
@@ -303,10 +343,11 @@ short *objectselect(class)
       unselect = True;
       class = -class;
    }
+   class &= areastruct.filter;	/* apply the selection filter */
 
    /* Loop through all elements found underneath the cursor */
 
-   for (curgen = objectdata->plist; curgen < ENDPART; curgen++) {
+   for (curgen = selobj->plist; curgen < selobj->plist + selobj->parts; curgen++) {
 
       /* when selection element to attach to, don't select the element */
       /* to be attached. . .					       */
@@ -347,47 +388,45 @@ short *objectselect(class)
          /*  point's y-position is the same for all bbox segments if the   */
          /*  pointer position is inside the bounding box.		   */
 
-         stval = sign((newboxpts[0].x - newboxpts[3].x) * (areastruct.save.y - 
-	    newboxpts[3].y) - (newboxpts[0].y - newboxpts[3].y) * (areastruct.save.x 
-	    - newboxpts[3].x));
-
-         for (j = 1; j < 4; j++) {
-	    stval += sign((newboxpts[j].x - newboxpts[j - 1].x) * (areastruct.save.y
-	    - newboxpts[j - 1].y) - (newboxpts[j].y - newboxpts[j - 1].y) * 
-	    (areastruct.save.x - newboxpts[j - 1].x));
-         }
-
-         if (abs(stval) == 4) selected = True;
+	 selected = test_insideness(areastruct.save.x, areastruct.save.y,
+		newboxpts);
       }
 
       else if ((*curgen)->type == (class & LABEL)) {
+	 labelptr curlab = TOLABEL(curgen);
 
          /* now look among the labels */
 
-	 short tmplength = ULength(TOLABEL(curgen)->string, 0.0, 0);
+	 TextExtents tmpext = ULength(curlab->string, 0.0, 0, NULL);
 
-         bboxpts[0].x = bboxpts[1].x = (TOLABEL(curgen)->justify & NOTLEFT ?
-	    (TOLABEL(curgen)->justify & RIGHT ? -tmplength : -tmplength / 2) : 0);
-	 bboxpts[2].x = bboxpts[3].x = bboxpts[0].x + tmplength;
-         bboxpts[0].y = bboxpts[3].y = (TOLABEL(curgen)->justify & NOTBOTTOM ?
-	    (TOLABEL(curgen)->justify & TOP ? -TEXTHEIGHT : -HALFHEIGHT) : 0);
-	 bboxpts[1].y = bboxpts[2].y = bboxpts[0].y + TEXTHEIGHT;
+         bboxpts[0].x = bboxpts[1].x = (curlab->justify & NOTLEFT ?
+	    (curlab->justify & RIGHT ? -tmpext.width : -tmpext.width / 2) : 0);
+	 bboxpts[2].x = bboxpts[3].x = bboxpts[0].x + tmpext.width;
+         bboxpts[0].y = bboxpts[3].y = (curlab->justify & NOTBOTTOM ?
+	    (curlab->justify & TOP ? -tmpext.ascent :
+	    -(tmpext.ascent + tmpext.base) / 2) : -tmpext.base)
+	    + tmpext.descent;
+	 bboxpts[1].y = bboxpts[2].y = bboxpts[0].y +
+		(tmpext.ascent - tmpext.descent);
 
-         UTransformPoints(bboxpts, newboxpts, 4, TOLABEL(curgen)->position,
-		TOLABEL(curgen)->scale, TOLABEL(curgen)->rotation);
+#ifdef SCHEMA
+	 if (curlab->pin) {
+	    int i;
+
+	    if (!areastruct.schemon) continue;
+	    for (i = 0; i < 4; i++) 
+	       pinadjust(curlab->justify, &(bboxpts[i].x), &(bboxpts[i].y), 1);
+	 }
+#endif
+
+         UTransformPoints(bboxpts, newboxpts, 4, curlab->position,
+		curlab->scale, curlab->rotation);
 
          /* check for point inside bounding box, as for objects */
 
-         stval = sign((newboxpts[0].x - newboxpts[3].x) * (areastruct.save.y -
-            newboxpts[3].y) - (newboxpts[0].y - newboxpts[3].y) * (areastruct.save.x 
-	    - newboxpts[3].x));
-
-         for (j = 1; j < 4; j++)
-            stval += sign((newboxpts[j].x - newboxpts[j - 1].x) * (areastruct.save.y
-            - newboxpts[j - 1].y) - (newboxpts[j].y - newboxpts[j - 1].y) *
-            (areastruct.save.x - newboxpts[j - 1].x));
-
-         if (abs(stval) == 4) selected = True;
+	 selected = test_insideness(areastruct.save.x, areastruct.save.y,
+		newboxpts);
+	 if (selected) textpos = textend = 0;
       }
 
 
@@ -399,10 +438,15 @@ short *objectselect(class)
 	 Boolean lselect;
 
 	 /* Accept path if any subcomponent of the path is accepted */
+	 /* Record which part was selected in the "subeditpart" variable */
 
  	 for (pathp = TOPATH(curgen)->plist; pathp < TOPATH(curgen)->plist +
 		TOPATH(curgen)->parts; pathp++)
-	    if (pathselect(pathp, SPLINE|ARC|POLYGON)) selected = True;
+	    if (pathselect(pathp, SPLINE|ARC|POLYGON)) {
+	       selected = True;
+	       areastruct.editsubpart = (short)(pathp - TOPATH(curgen)->plist);
+	       break;
+	    }
       }
 
       /* Check among polygons, arcs, and curves */
@@ -413,7 +457,7 @@ short *objectselect(class)
 
       for (newselect = areastruct.selectlist; newselect <
              areastruct.selectlist + areastruct.selects; newselect++) 
-         if (*newselect == (short)(curgen - objectdata->plist)) {
+         if (*newselect == (short)(curgen - selobj->plist)) {
 	    if (!unselect) selected = False;
 	    break;
 	 }
@@ -433,7 +477,7 @@ short *objectselect(class)
          newselect = foundlist + found;
          found++;
      
-         if (selected) *newselect = (short)(curgen - objectdata->plist);
+         if (selected) *newselect = (short)(curgen - selobj->plist);
       }
 
       /* mechanism for unselecting elements under the cursor */
@@ -442,8 +486,8 @@ short *objectselect(class)
          short *desel;
 
          XSetFunction(dpy, areastruct.gc, GXcopy);
-         XTopSetForeground(SELTOCOLOR(newselect));
-         easydraw(*newselect, DEFAULTCOLOR);
+         XTopSetForeground(GSELTOCOLOR(selobj, newselect));
+         geneasydraw(*newselect, DEFAULTCOLOR, selobj);
          areastruct.selects--;
          for (desel = newselect; desel < areastruct.selectlist +
 	        areastruct.selects; desel++)
@@ -465,14 +509,14 @@ short *objectselect(class)
 	 newselect = allocselect();
 	 *newselect = *foundlist;
 	 free(foundlist);
-	 drawselected(newselect);
+	 gendrawselected(newselect, selobj);
 	 break;
 
       default: {
 	 XSetFunction(dpy, areastruct.gc, GXcopy);
 	 for (foundtmp = foundlist; foundtmp < foundlist + found; foundtmp++) {
             XSetForeground(dpy, areastruct.gc, QUERYCOLOR);
-            easydraw(*foundtmp, DOFORALL);
+            geneasydraw(*foundtmp, DOFORALL, selobj);
 	    sprintf(_STR, "Click to accept/reject: %d of %d",
 		(int)(foundtmp - foundlist + 1), found);
 	    Wprintf(_STR);
@@ -485,14 +529,14 @@ short *objectselect(class)
 	       else if (event.type == ButtonPress) break;
 	    }
 	    if (buttonevent->button == Button3) {
-               XTopSetForeground(SELTOCOLOR(foundtmp));
-               easydraw(*foundtmp, DEFAULTCOLOR);
+               XTopSetForeground(GSELTOCOLOR(selobj, foundtmp));
+               geneasydraw(*foundtmp, DEFAULTCOLOR, selobj);
 	    }
 	    else {
 	       newselect = allocselect();
                *newselect = *foundtmp;
                XSetForeground(dpy, areastruct.gc, SELECTCOLOR);
-               easydraw(*newselect, DOFORALL);
+               geneasydraw(*newselect, DOFORALL, selobj);
 	    }
          }
       }  break;
@@ -525,12 +569,12 @@ short *objectselect(class)
       XSetForeground(dpy, areastruct.gc, BACKGROUND);
       for (nselect = areastruct.selectlist; nselect < areastruct.selectlist
 	   + areastruct.selects; nselect++)
-         easydraw(*nselect, DOFORALL);
+         geneasydraw(*nselect, DOFORALL, selobj);
       XSetFunction(dpy, areastruct.gc, GXxor);
       XSetForeground(dpy, areastruct.gc, SELECTCOLOR ^ BACKGROUND);
       for (nselect = areastruct.selectlist; nselect < areastruct.selectlist
 	   + areastruct.selects; nselect++)
-         easydraw(*nselect, DOFORALL);
+         geneasydraw(*nselect, DOFORALL, selobj);
    }
 
    setoptionmenu();
@@ -623,27 +667,90 @@ void selectarea()
       areastruct.save.x = tmp;
    }
  
+   textpos = textend = 0;
    for (curgen = objectdata->plist; curgen < objectdata->plist
 	+ objectdata->parts; curgen++) {
 
+      /* apply the selection filter */
+      if (!((*curgen)->type & areastruct.filter)) continue;
+
       switch((*curgen)->type) {
-	 case(OBJECT):
-            /* check for instance of object inside area box */
-            selected = (TOOBJINST(curgen)->position.x < areastruct.save.x) &&
+	case(OBJECT):
+           /* check for instance of object inside area box */
+           selected = (TOOBJINST(curgen)->position.x < areastruct.save.x) &&
 		(TOOBJINST(curgen)->position.x > areastruct.origin.x) &&
 		(TOOBJINST(curgen)->position.y < areastruct.save.y) &&
 		(TOOBJINST(curgen)->position.y > areastruct.origin.y);
-	    break;
-
-        case(LABEL):
-   	   /* check position point of each label */
-
-      	   selected = (TOLABEL(curgen)->position.x < areastruct.save.x) &&
-		(TOLABEL(curgen)->position.x > areastruct.origin.x) &&
-		(TOLABEL(curgen)->position.y < areastruct.save.y) &&
-		(TOLABEL(curgen)->position.y > areastruct.origin.y);
 	   break;
 
+        case(LABEL): {
+	   XPoint bboxpts[4], newboxpts[4], adj;
+	   labelptr slab = TOLABEL(curgen);
+	   short j, state, isect, tmpl1, tmpl2;
+	   TextExtents tmpext;
+
+	   selected = False;
+
+	   /* create a 4-point polygon out of the select box information */
+	   bboxpts[0].x = bboxpts[3].x = areastruct.origin.x;
+	   bboxpts[0].y = bboxpts[1].y = areastruct.origin.y;
+	   bboxpts[1].x = bboxpts[2].x = areastruct.save.x;
+	   bboxpts[2].y = bboxpts[3].y = areastruct.save.y;
+
+   	   /* translate select box into the coordinate system of the label */
+	   InvTransformPoints(bboxpts, newboxpts, 4, slab->position,
+		  slab->scale, slab->rotation);
+
+#ifdef SCHEMA
+	   if (slab->pin) {
+	      if (!areastruct.schemon) continue;
+	      for (j = 0; j < 4; j++) 
+		 pinadjust(slab->justify, &(newboxpts[j].x),
+			&(newboxpts[j].y), -1);
+	   }
+#endif
+	   tmpext = ULength(slab->string, 0.0, 0, NULL);
+	   adj.x = (slab->justify & NOTLEFT ? (slab->justify & RIGHT ? 
+			tmpext.width : tmpext.width >> 1) : 0);
+	   adj.y = (slab->justify & NOTBOTTOM ? (slab->justify & TOP ? 
+			tmpext.ascent : (tmpext.ascent + tmpext.base) >> 1)
+			: tmpext.base);
+	   
+	   /* Label selection:  For each character in the label string, */
+	   /* do an insideness test with the select box.		*/
+
+	   state = tmpl2 = 0;
+	   for (j = 2; j < strlen(slab->string); j++) {
+	      if (*(slab->string + j) == TEXT_ESC
+			&& *(slab->string + j + 1) == RETURN)
+		 tmpl2 = 0;
+	      tmpl1 = tmpl2;
+	      tmpext = ULength(slab->string, 0.0, j + 1, NULL);
+	      tmpl2 = tmpext.width;
+	      isect = test_insideness(((tmpl1 + tmpl2) >> 1) - adj.x,
+			(tmpext.base + (BASELINE >> 1)) - adj.y, newboxpts);
+
+	      /* tiny state machine */
+	      if (state == 0) {
+		 if (isect) {
+		    state = 1;
+		    selected = True;
+		    textend = j;
+		    if ((textend > 1) && (*(slab->string + textend - 1) == TEXT_ESC))
+		       textend--;
+		 }
+	      }
+	      else {
+		 if (!isect) {
+		    textpos = j;
+		    state = 2;
+		    break;
+		 }
+	      }
+	   }
+	   if (state == 1) textpos = j;   /* selection goes to end of string */
+	   } break;
+	   
 	case(PATH):
 	   /* check position point of each subpart of the path */
 
@@ -666,7 +773,7 @@ void selectarea()
               areastruct.selectlist + areastruct.selects; newselect++)
             if (*newselect == (short)(curgen - objectdata->plist))
                   selected = False;
-	
+
       /* add to list of selections */
 
       if (selected) {
@@ -676,6 +783,13 @@ void selectarea()
    }
    setoptionmenu();
 
+   /* if none or > 1 label has been selected, cancel any textpos placement */
+
+   if (!checkselect(LABEL) || areastruct.selects != 1 ||
+	(areastruct.selects == 1 && SELECTTYPE(areastruct.selectlist) != LABEL)) {
+      textpos = textend = 0;
+   }
+
    /* Drawing of selected objects will take place when drawarea() is */
    /* executed after the button release.			     */
 }
@@ -710,6 +824,8 @@ void deselect(w, clientdata, calldata)
    setoptionmenu();
 }
 
+/*------------------------------------------*/
+
 void objectdeselect()
 {
    short *lastselect;
@@ -734,6 +850,60 @@ void objectdeselect()
    setdefaultfontmarks();
 }
 
+/*------------------------------------------------------------------------*/
+/* Select the nearest polygon, searching down the hierarchy if necessary. */
+/* Return pointers to the polygon and the object containing the polygon.  */
+/*------------------------------------------------------------------------*/
+
+objectptr recurselect(polyptr *polyreturn, objectptr selobj)
+{
+   short *selectobj;
+   objinstptr selinst;
+   objectptr selreturn;
+   XPoint savesave, tmppt;
+
+   if (areastruct.selects > 0) {
+      areastruct.selects = 0;
+      free(areastruct.selectlist);
+   }
+
+   if ((selectobj = selectobject(POLYGON, selobj)) != NULL) {
+      *polyreturn = TOPOLY(selobj->plist + (*selectobj));
+      areastruct.selects = 0;
+      free(areastruct.selectlist);
+      /* printf("polygon found in object %s\n", selobj->name); */
+      selreturn = selobj;
+   }
+   else if ((selectobj = selectobject(OBJECT, selobj)) != NULL) {
+      if ((*(selobj->plist + *selectobj))->type == OBJECT) {
+         selinst = TOOBJINST(selobj->plist + (*selectobj));	
+         /* Translate areastruct.save into object's own coordinate system */
+         savesave.x = areastruct.save.x;
+         savesave.y = areastruct.save.y;
+         InvTransformPoints(&areastruct.save, &tmppt, 1, selinst->position,
+		selinst->scale, selinst->rotation);
+         areastruct.save.x = tmppt.x;
+         areastruct.save.y = tmppt.y;
+         /* printf("objinst %s found in object %s; searching recursively\n",
+		selinst->thisobject->name, selobj->name); */
+         /* printf("cursor position originally (%d, %d); in new object is (%d, %d)\n",
+		savesave.x, savesave.y,
+		areastruct.save.x, areastruct.save.y); */
+	 UPushCTM();
+	 UPreMultCTM(DCTM, selinst->position, selinst->scale, selinst->rotation);
+         selreturn = recurselect(polyreturn, selinst->thisobject);
+         UPopCTM();
+         areastruct.save.x = savesave.x;
+         areastruct.save.y = savesave.y;
+      }
+   }
+   else {
+      /* printf("nothing found in object %s\n", selobj->name); */
+      selreturn = (objectptr)NULL;
+   }
+   return selreturn;
+}
+
 /*----------------------------------*/
 /* Start drawing a select area box. */
 /*----------------------------------*/
@@ -766,7 +936,7 @@ XtEventHandler trackselarea(w, clientdata, event)
 
    XQueryPointer(dpy, win, &nullwin, &nullwin, &nullint, &nullint, &xpos, &ypos,
 	&nullui);
-   snap(xpos, ypos, &newpos);
+   window_to_user(xpos, ypos, &newpos);
    if (newpos.x == areastruct.save.x && newpos.y == areastruct.save.y) return;
 
    UDrawBox(areastruct.origin, areastruct.save);
